#ifndef LegoGenHH
#define LegoGenHH

/* Fills the allowed regions of dest with Lego1 and Lego2.
 * dest is 16x16. allowed is 32x16.
 *
 * Lego1 and Lego2 are assumed to be pointers into 3x2 arrays of bytes,
 * where the middle part is optional.
 */
void PutLego(unsigned char* dest, unsigned x,unsigned y, unsigned width,
             const unsigned char Lego[2][3])
{
    for(unsigned ny=0; ny<2; ++ny)
        for(unsigned nx=0; nx<width; ++nx)
        {
            unsigned p = nx==0 ? 0 : nx==(width-1) ? 2 : 1;
            char c = Lego[ny][p];
            dest[(y+ny)*16+(x+nx)] = c;
        }
}
               
unsigned AttemptPlaceLego(unsigned char* dest,
                          const char* allowed, char allowedchar,
                          const unsigned char Lego1[2][3],
                          const unsigned char Lego2[2][3],
                          unsigned x,unsigned y,
                          bool may_overwrite)
{
    unsigned width;
    
    /* Verify how wide a lego can be put in this position */
    for(width=0; width<3; ++width)
    {
        for(unsigned y2=0; y2<2; ++y2)
        {
            unsigned yy=y+y2;
            unsigned xx=x+width;
            if((!may_overwrite && dest[yy*16+xx] != 0)
            || allowed[yy*32+xx] != allowedchar) goto nok;
        }
        continue;
    nok: break;
    }
    
    if(width < 2)
    {
        /* No lego could be placed */
        return 0;
    }

    /* Put a randomly selected lego to the position */
    if(rand()%2)
        PutLego(dest, x,y, width, Lego1);
    else
        PutLego(dest, x,y, width, Lego2);
    return width;
}

void LegoFill(unsigned char* dest,
              const char* allowed, char allowedchar,
              const unsigned char Lego1[2][3],
              const unsigned char Lego2[2][3]
             )
{
    unsigned char result[16*16];
    
    // First, paint the background.
    for(unsigned y=0; y<16; y+=2)
        for(unsigned x=0; x<16; )
        {
            unsigned width = 2 + (rand()%2);
            if(x == 13) width = 3; else if(x == 14) width = 2;
            
            if(rand()%2)
                PutLego(result, x,y, width, Lego1);
            else
                PutLego(result, x,y, width, Lego2);
            x += width;
        }
    
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            if(allowed[y*32+x] == allowedchar)
                dest[y*16+x] = 0;

    // Now attempt to place them in a more intelligent way...
    
    // Find each corner.
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            /* Is there an edge on each direction? */
            bool up    = y==0 || allowed[(y-1)*32+x]!=allowedchar;
            bool left  = x==0 || allowed[y*32+(x-1)]!=allowedchar;
            bool down  = y==15 || allowed[(y+1)*32+x]!=allowedchar;
            bool right = x==15 || allowed[y*32+(x+1)]!=allowedchar;
            
            if(up != down && left != right)
            {
                unsigned legox = x, legoy = y;
                if(down)  legoy -= 1; /* If the edge is below, move 1 up */
                if(right) legox -= 1; /* If the edge is on right, move 1 left */
                
                /* Allow overlapping. */
                unsigned width =
                    AttemptPlaceLego(dest,allowed,allowedchar,
                                     Lego1,Lego2,
                                     legox,legoy,
                                     true);
                
                if(up && left)
                {
                    legox += width;
                    unsigned tmpy = legoy;
                    while(tmpy+2 < 16)
                    {
                        unsigned tmpx = legox;
                        /* Go backwards as much as there's room */
                        
                        while(tmpx > 0 && allowed[(tmpy+0)*32+tmpx-1]==allowedchar
                                       && allowed[(tmpy+1)*32+tmpx-1]==allowedchar)
                        {
                            --tmpx;
                        }
                        
                        for(;;)
                        {
                            if(tmpx+2 > 16) break;
                            width = AttemptPlaceLego(dest, allowed, allowedchar,
                                                     Lego1,Lego2, tmpx,tmpy, false);
                            if(!width) width=1;
                            tmpx += width;
                        }
                        tmpy += 2;
                    }
                }
            }
        }

    // Place random legoes
    for(unsigned n=0; n<1000; ++n)
    {
        unsigned x = (rand()%(16-2));
        unsigned y = (rand()%(15-2));
        
        /* Only where they fit without overlapping. */        
        AttemptPlaceLego(dest,allowed,allowedchar, Lego1,Lego2, x,y, false);
    }
    
    // Fill the background where nothing else exists
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            if(allowed[y*32+x] == allowedchar
            && dest[y*16+x]==0)
                dest[y*16+x] = result[y*16+x];
}

#endif
